################################################################################
# Key Tools
#
# This module contains utilities that help with parsing and generating keys
# that are used with Disid. Keys are not required for use but provide an
# increased level of obfuscation to the casual observer. They function as
# substitution ciphers for each letter of the ID which helps disguise simple
# patterns found in the ID. Adding a key does not add any meaningful level of
# security to the IDs against a dedicated attacker however. Disid is not meant
# to be used in and should not be used in any situation where the original
# integer ID must not be shared with the user
################################################################################
from typing import List, Dict
import os

key_chunk_size = 37


################################################################################
# generate_key
#
# Generates a new key to be used with Disid. Returns a bytes object
# containing a valid 190 byte long key which can be used for a 5 character long
# disguised value unless the size parameter is passed in, in which case a
# different length key will be returned instead.
################################################################################
def generate_key(size: int = 5) -> bytes:
    return b''.join([_generate_key_chunk() for _ in range(size)])


################################################################################
# convert_key_to_character_list
#
# Takes in a key, like one generated from generate_key(), and transforms it
# into an array of character lists that can be used to generate the Disid.
################################################################################
def convert_key_to_character_lists(key: bytes) -> List[str]:
    if len(key) % key_chunk_size != 0:
        raise ValueError("Key must be a multiple of key_chunk_size bytes")

    index = 0
    memkey = memoryview(key)

    character_lists = []

    while index + key_chunk_size <= len(memkey):
        character_lists.append(
            _character_list_from_index(int.from_bytes(memkey[index:index + key_chunk_size], 'big'))
        )
        index += key_chunk_size

    return character_lists


################################################################################
# generate_key_chunk
#
# Generates a chunk of the key, representative of a single permutation of one
# of the characters in the id.
################################################################################
def _generate_key_chunk() -> bytes:
    key_chunk = os.urandom(key_chunk_size)

    key_int = int.from_bytes(key_chunk, 'big')

    # A "slow and dumb" way to make sure we have a valid chunk of bytes, but
    # there is only a 0.3% chance of that happening [1-64!/2^(37*8) = 0.00336]
    # so it will basically have no impact on performance making this pretty
    # fast on average.
    while key_int >= _factorial(64):
        key_chunk = os.urandom(key_chunk_size)
        key_int = int.from_bytes(key_chunk, 'big')

    return key_chunk


################################################################################
# character_list_from_index
#
# This function takes in an index of one of the 64! permutations of the 64
# characters, and returns a string with the characters in that particular
# permutation. The characters list is the base64 character set except for the
# two symbols are `-` and `_` instead of `+` and `/`.
################################################################################
characters = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_-"


def _character_list_from_index(index: int) -> str:
    permutation: List[int] = _permutation_from_index(64, index)
    return "".join([characters[i] for i in permutation])


################################################################################
# permutation_from_index
#
# Returns an array of length `length` with one of the possible permutations of
# integers from `0` to `length-1`. The index value can be any value between and
# including 0 and factorial(length)-1. This operation is deterministic so the
# same combination of  length and index will always return the same
# permutation. This function does not calculate all the permutations, reducing
# the complexity to  O(LOG(length!)), mostly due to handling large integer
# values of `index`.
################################################################################
def _permutation_from_index(length: int, index: int) -> List[int]:
    # Sanity check that the permutation index is a valid index.
    if index >= _factorial(length) or index < 0:
        raise ValueError("Index out of bounds")

    # Transform the index into an array of possible options with the maximum
    # value of each element being the number of remaining options available for
    # values at that index. The maximum values would be:
    #
    #     [length-1, length-2, length-3, length-4, ... , 3, 2, 1, 0]
    #
    # The last element will always be 0 because the final element only ever has
    # a single option remaining, leaving this 0 in the array makes the next
    # part of this function simpler.
    permutation: List[int] = []
    for i in range(length):
        permutation.append(index // _factorial(length - i - 1))
        index = index % _factorial(length - i - 1)

    # Convert the permutation array of decreasing maximum values into a non
    # repeating array containing all values between and including 0 and
    # length-1 by removing previously chosen values from a list of all possible
    # values before the next element chooses a value.
    options = [option for option in range(length)]
    value: List[int] = []
    for i in permutation:
        value.append(options.pop(i))
    return value


################################################################################
# factorial
#
# A memoized factorial function.
################################################################################
_factorials: Dict[int, int] = {0: 1, 1: 1}


def _factorial(i: int) -> int:
    global _factorials

    if i not in _factorials:
        _factorials[i] = _factorial(i - 1) * i

    return _factorials[i]



################################################################################
# Disid
#
# Disid is a tool for converting an integer into a psuedo random id. Its
# purpose is to disguise a lineraly increasing numerical value so that it
# apears to be a random string of letters and numbers. This is not
# cryptographically secure and with enough known consecutive inputs a dedicated
# attacker can easily rebuild the keys. This should only be used to enhance
# the style of a user facing element that could otherwise contain the raw
# integer value.
#
# Every ID and integer have a 1 to 1 mapping, meaning that every ID represents
# a single integer and ever integer represents a single ID.
################################################################################
from typing import List


################################################################################
# uint_to_id_v3
#
# Converts a positive integer into a V3 id string.
################################################################################
def uint_to_id_v3(integer: int, digit_order: List[str]) -> str:
    length = len(digit_order)

    base_64 = uint_to_base64(integer)
    base_64 = pad_zeros(base_64, length)

    for late_index in reversed(range(1, length)):
        for early_index in range(0, late_index):
            base_64[late_index] = base_64[late_index] + (base_64[early_index] + 1) * (late_index - early_index)
        base_64[late_index] = base_64[late_index] % 64

    identifier = ""
    for i in range(length):
        identifier += digit_order[i][base_64[i]]

    return identifier


################################################################################
# id_v2_to_uint
#
# Depricated. Use uint_to_id_v3() and id_v3_to_uint() instead.
# Converts a v2 id value into a positive integer.
################################################################################
def id_v3_to_uint(identifier: str, digit_order: List[str]) -> int:
    length = len(identifier)

    base_64 = []
    for i in range(length):
        base_64.append(digit_order[i].index(identifier[i]))

    for early_index in range(0, length):
        base_64[early_index] = base_64[early_index] % 64
        for late_index in range(early_index + 1, length):
            base_64[late_index] = base_64[late_index] - (base_64[early_index] + 1) * (late_index - early_index)

    # Decode the values into an int
    modifier = 1
    integer = 0
    for digit in base_64:
        integer += digit * modifier
        modifier *= 64

    return integer


################################################################################
# uint_to_id_v2
#
# Depricated. Use unit_to_id_v3() instead. V2 suffers from an issue where
# characters are more likely to show up in a pattern the farther into the id
# you get.
# Converts a positive integer into a V2 id string.
################################################################################
def uint_to_id_v2(integer: int, digit_order: List[str]) -> str:
    base_64 = uint_to_base64(integer)
    base_64 = pad_zeros(base_64, len(digit_order))

    offset = base_64[0] + 1
    for i in range(1, len(digit_order)):
        base_64[i] = (base_64[i] + offset) % 64
        offset += base_64[i]

    identifier = ""
    for i in range(len(digit_order)):
        identifier += digit_order[i][base_64[i]]

    return identifier


################################################################################
# id_v2_to_uint
#
# Depricated. Use uint_to_id_v3() and id_v3_to_uint() instead.
# Converts a v2 id value into a positive integer.
################################################################################
def id_v2_to_uint(identifier: str, digit_order: List[str]) -> int:
    offset_base_64 = []
    for i in range(5):
        offset_base_64.append(digit_order[i].index(identifier[i]))

    base_64 = [offset_base_64[0]]
    offset = offset_base_64[0] + 1
    for i in offset_base_64[1:]:
        base_64.append((i - offset) % 64)
        offset += i

    modifier = 1

    integer = 0

    for digit in base_64:
        integer += digit * modifier
        modifier *= 64

    return integer


################################################################################
# uint_to_id_v1
#
# Depricated. Use unit_to_id_v3() instead. V1 suffers from an issue wher all
# but the first character get repeated nearly every 63 characters.
# Converts a positive integer into a V1 id string.
################################################################################
def uint_to_id_v1(integer: int, digit_order: List[str]) -> str:
    base_64 = uint_to_base64(integer)
    base_64 = pad_zeros(base_64, len(digit_order))

    for i in range(1, len(digit_order)):
        base_64[i] = (base_64[i] + base_64[i - 1]) % 64

    identifier = ""
    for i in range(len(digit_order)):
        identifier += digit_order[i][base_64[i]]

    return identifier


################################################################################
# id_v1_to_uint
#
# Depricated. Use uint_to_id_v3() and id_v3_to_uint() instead.
# Converts a v1 id value into a positive integer.
################################################################################
def id_v1_to_uint(identifier: str, digit_order: List[str]) -> int:
    base_64 = []
    for i in range(5):
        base_64.append(digit_order[i].index(identifier[i]))

    for i in range(4, 0, -1):
        base_64[i] = (base_64[i] - base_64[i - 1]) % 64

    modifier = 1

    integer = 0

    for digit in base_64:
        integer += digit * modifier
        modifier *= 64

    return integer


################################################################################
# int_to_base64
#
# Converts a regular unsigned integer to a list of base64 digits. The first
# element in the returned list is the least significant digit. If a negative
# number is received it will error.
#
# TODO: Because this is just base64 now instead of any base we can take
# advantage of bit-slicing and bitwise operators to make this faster.
################################################################################
def uint_to_base64(x: int) -> List[int]:
    if x < 0:
        raise ValueError("Cannot turn negative int into base64.")
    elif x == 0:
        return [0]

    digits = []

    while x:
        digits.append(int(x % 64))
        x = int(x / 64)

    return digits


################################################################################
# pad_zeros
#
# A helper function to add 0's to the end of the digit list to make it five
# elements long.
################################################################################
def pad_zeros(array: List[int], length: int = 5) -> List[int]:
    delta_length = length - len(array)

    for _ in range(delta_length):
        array.append(0)

    return array
